@@ -50,6 +50,7 @@
 #include <QLocale>
 #include <QIcon>
 #include <QColor>
+#include <QRegularExpression>
 
 #include <gpgme++/gpgmepp_version.h>
 #if GPGMEPP_VERSION > 0x10900
@@ -258,6 +259,19 @@ QString format_keytype(const Key &key)
     }
 }
 
+QString format_subkeytype(const Subkey &subkey)
+{
+    const auto algo = subkey.publicKeyAlgorithm();
+
+    if (algo == Subkey::AlgoECC ||
+        algo == Subkey::AlgoECDSA ||
+        algo == Subkey::AlgoECDH ||
+        algo == Subkey::AlgoEDDSA) {
+        return QString::fromStdString(subkey.algoName());
+    }
+    return i18n("%1-bit %2", subkey.length(), QLatin1String(subkey.publicKeyAlgorithmAsString()));
+}
+
 QString format_keyusage(const Key &key)
 {
     QStringList capabilities;
@@ -280,6 +294,28 @@ QString format_keyusage(const Key &key)
     return capabilities.join(QStringLiteral(", "));
 }
 
+QString format_subkeyusage(const Subkey &subkey)
+{
+    QStringList capabilities;
+    if (subkey.canSign()) {
+        if (subkey.isQualified()) {
+            capabilities.push_back(i18n("Signing (Qualified)"));
+        } else {
+            capabilities.push_back(i18n("Signing"));
+        }
+    }
+    if (subkey.canEncrypt()) {
+        capabilities.push_back(i18n("Encryption"));
+    }
+    if (subkey.canCertify()) {
+        capabilities.push_back(i18n("Certifying User-IDs"));
+    }
+    if (subkey.canAuthenticate()) {
+        capabilities.push_back(i18n("SSH Authentication"));
+    }
+    return capabilities.join(QStringLiteral(", "));
+}
+
 static QString time_t2string(time_t t)
 {
     QDateTime dt;
@@ -395,6 +431,40 @@ QString Formatting::toolTip(const Key &key, int flags)
             result += format_row(i18n("Stored"), i18nc("stored...", "on this computer"));
         }
     }
+    if (flags & Subkeys) {
+        for (const auto &sub: key.subkeys()) {
+            result += QLatin1String("<hr/>");
+            result += format_row(i18n("Subkey"), sub.fingerprint());
+            if (sub.isRevoked()) {
+                result += format_row(i18n("Status"), i18n("Revoked"));
+            } else if (sub.isExpired()) {
+                result += format_row(i18n("Status"), i18n("Expired"));
+            }
+            if (flags & ExpiryDates) {
+                result += format_row(i18n("Created"), time_t2string(sub.creationTime()));
+
+                if (key.isExpired()) {
+                    result += format_row(i18n("Expired"), time_t2string(sub.expirationTime()));
+                } else if (!subkey.neverExpires()) {
+                    result += format_row(i18n("Expires"), time_t2string(sub.expirationTime()));
+                }
+            }
+
+            if (flags & CertificateType) {
+                result += format_row(i18n("Type"), format_subkeytype(sub));
+            }
+            if (flags & CertificateUsage) {
+                result += format_row(i18n("Usage"), format_subkeyusage(sub));
+            }
+            if (flags & StorageLocation) {
+                if (const char *card = sub.cardSerialNumber()) {
+                    result += format_row(i18n("Stored"), i18nc("stored...", "on SmartCard with serial no. %1", QString::fromUtf8(card)));
+                } else {
+                    result += format_row(i18n("Stored"), i18nc("stored...", "on this computer"));
+                }
+            }
+        }
+    }
     result += QLatin1String("</table>");
 
     return result;
@@ -641,10 +711,8 @@ QIcon Formatting::validityIcon(const UserID::Signature &sig)
         return QIcon::fromTheme(QStringLiteral("emblem-information"));
     case UserID::Signature::NoPublicKey:
         return QIcon::fromTheme(QStringLiteral("emblem-question"));
-    default:
-        // should not happen
-        return QIcon();
     }
+    return QIcon();
 }
 
 QString Formatting::formatKeyLink(const Key &key)
@@ -932,6 +1000,17 @@ QString Formatting::complianceStringShort(const GpgME::Key &key)
         }
         return i18nc("As in all user IDs are valid.", "certified");
     }
+    if (key.isExpired()) {
+        return i18n("expired");
+    }
 
     return i18nc("As in not all user IDs are valid.", "not certified");
 }
+
+QString Formatting::prettyID(const char *id)
+{
+    if (!id) {
+        return QString();
+    }
+    return QString::fromLatin1(id).toUpper().replace(QRegularExpression(QStringLiteral("(....)")), QStringLiteral("\\1 ")).trimmed();
+}
